/***************************************************************************
                          databaseaccess.cpp  -  description
                             -------------------
    begin                : Fr Nov 5 2004
    copyright            : (C) 2004 by Marcel Mulch
    email                : marcel@mulch-online.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "databaseaccess.h"
#include "globals.h"

#include <iostream>
#include <fstream>
using namespace std;

#include <qstring.h>
#include <qdatetime.h>
#include <qfile.h>
#include <qfileinfo.h>

#include <gwenhywfar/text.h>

DatabaseAccess::DatabaseAccess(Pointer<AqMoneyAPI> api){
  _api = api;
}
DatabaseAccess::~DatabaseAccess(){
 qDebug("Destructor DatabaseAccess");
}



/**  This method creates a new User configuration. Like most of the code
  *  hobit consists of this method was also inspired by AqMoney.
  *
  * @author Marcel Mulch
  * @date 20.10.2004
  *
  * @param mediumname  the name of the security medium. this might be the
  *                    path to the keyfile if you are using an rsa-medium
  *
  * @param mediumtype  the type of the medium. OpenHBCI (and hbci in general)
  *                    differs between 2 types of security mediums. The
  *                    first one is RSA-File. The other one is DDV-Chipcard
  *                    With this parameter you can tell HoBIT (and OpenHBCI)
  *                    which type of medium you are using. <br> <br>
  *
  *                    If you are using RDH-File:   "RDHFile" <br>
  *                    If you are using DDV-Card:   "???"     <br>
  *
  * @param readonly    You can decide if the security medium should be set to
  *                    readonly. If you want so set readonly to true
  *
  * @param country     This is the countrycode. It is needed to tell HBCI the
  *                    location of your bank. This parameter should be 280 if
  *                    you are trying to connect to a german bank.
  *
  * @param instid      The institute code. Every bank has a special code, what
  *                    is unique and identifies it. In Germany this bank code
  *                    is just called Bankleitzahl.
  *
  * @param userid      In HBCI users have an Identification code too. Because
  *                    Users get this code from their bank I do not have to
  *                    explain it. Look into the HBCI-Specification if unsure.
  *
  * @param customerid  In HBCI customers have an Identification code too. Because
  *                    Customers get this code from their bank I do not have to
  *                    explain it. Look into the HBCI-Specification if unsure
  *                    In most cases the customerid will be equal to the user id.
  *
  *
  * @param server      this might be the url or the ip-adress of the bank server
  */
  Error DatabaseAccess :: createUser( const string &mediumname,
                                      const string &mediumType,
                                      bool readonly,
                                      int country,
                                      const string &instid,
                                      const string &userid,
                                      const string &customerid,
                                      const string &server){

  Pointer<Bank> bank;
  Pointer<User> user;
  Pointer<Customer> cust;
  Pointer<Medium> medium;
  list<Pointer<User> > users;
  Error err;
  string ucid;

   if (country==0 ||
      instid.empty() ||
      userid.empty() ||
      customerid.empty() ||
      server.empty())
    return HBCI::Error("createUser()",
		       ERROR_LEVEL_NORMAL,
		       0,
		       ERROR_ADVISE_DONTKNOW,
		       "empty arguments");

  // create user and medium
  try {
    // check if the bank exists

    bank = _api.ref().findBank(country,instid);
    if (!bank.isValid()) {
      Pointer<MessageEngine> me;
      Pointer<Transporter> tr;

      // need to create it
      cerr << "Creating bank\n";
      bank= _api.ref().bankFactory(country,
                                   instid,
                                   server);

      cerr << "Setting HBCI-Version (stdvalue)\n";
      bank.ref().setHbciVersion(AQMONEY_DEFAULT_HBCI_VERSION);

      cerr << "Creating MessageEngine\n";
      me=_api.ref().engineFactory(country,
                                  instid,
                                  AQMONEY_DEFAULT_HBCI_VERSION,
                                  AQMONEY_DEFAULT_MSG_ENGINE);

      cerr << "creating default transporter\n";
      tr=_api.ref().transporterFactory(server,
                                       AQMONEY_DEFAULT_CONN_PORT,
                                       AQMONEY_DEFAULT_CONN_TIMEOUT,
                                       AQMONEY_DEFAULT_CONN_TYPE);

        bank.ref().setMessageEngine(me);
        bank.ref().setTransporter(tr);
        _api.ref().addBank(bank);
    }
    else {
      cerr << "Bank does already exist\n";
    }

    // check if user exists
     user=bank.ref().findUser(userid);
      if (!user.isValid()) {
        // need to create him
        cerr << "Now creating user\n";
        // check if medium exists
        medium= _api.ref().findMedium(mediumname);
        if (!medium.isValid()){

     //  need to create the medium
	  medium = _api.ref().createNewMedium(  mediumType,
                                                readonly,
                                                country,
                                                instid,
                                                userid,
                                                mediumname);
      }
      // then the user himself
       user=  _api.ref().userFactory(bank,
                                     medium,
                                     userid);
      // add customer to the system
      bank.ref().addUser(user);
      // done
     }
    else cerr << "\nUser already exists\n";
   


    // create customer, if given
      cust=user.ref().findCustomer(customerid);
     if (!cust.isValid()) {
        cerr << "Creating customer\n";
        cust = _api.ref().customerFactory(user, customerid);
        user.ref().addCustomer(cust);
      }
      else  cerr << "Customer does already exist\n";
      
  } // try
    catch (HBCI::Error lerr) {
      err=lerr;
      cout << err.errorString().c_str();
    }

  // after all: save the environment
    if (!err.isOk()){
       cerr << "There was an error during creating user\n";
       cerr << err.errorString().c_str();
    }

  return err;
}



/**  This method creates a new User configuration imported from a ddv card.
  *  Like most of the code hobit consists of this method was also inspired 
  *  by AqMoney.
  */
 Error DatabaseAccess::importUserDDV( bool readonly,
                                      int country,
                                      const string &instid,
                                      const string &userid,
                                      const string &customerid,
                                      const string &server){
  string tmp;
  Pointer<Medium> medium;
  int i;
  int j;
  int lcountry;
  string linstid;
  string luserid;
  string lserver;
  string mediumname;
  string mtype = "DDVCard2";
  Error err;

  // now import the contexts from the medium
  medium=_api.ref().mediumFactory(mediumname,mtype);
  try {
    i=1;
    j=0;
    err=medium.ref().mountMedium();
    if (!err.isOk())
      return HBCI::Error("importUser",err);
    while(1) {
      // get context
      err=medium.ref().getContext(i,
				  lcountry,
				  linstid,
				  luserid,
				  lserver);
      if (!err.isOk())
	break;
      // add context, if it matches the criteria
      if (
	  (!country || lcountry==country) &&
          (-1!=GWEN_Text_ComparePattern(linstid.c_str(),
                                        instid.c_str(),
                                        0) || instid.empty())
         ) {
        mediumname=medium.ref().mediumName();
         err=createUser( mediumname,
                         mtype,
                         readonly,
                         lcountry,
                         linstid,
                         userid.empty()?luserid:userid,
                         customerid,
                         lserver.empty()?server:lserver);

	if (!err.isOk()){
	  fprintf(stderr,I18N("Error: %s\n"),
		  err.errorString().c_str());
		  return err;
	   }
	j++;
      }
      i++;
    } // while
    medium.ref().unmountMedium();
  } // try
  catch (Error lerr) {
    fprintf(stderr,"Exception: %s\n",
	    lerr.errorString().c_str());
    	    return HBCI::Error("importUser",err);
  }
  if (!j) {
    fprintf(stderr,I18N("No user added.\n"));
    return HBCI::Error("importUser()",
		       ERROR_LEVEL_NORMAL,
		       0,
		       ERROR_ADVISE_ABORT,
		       "no user added");
  }

  return HBCI::Error();
 }        


/** Returns a std::list which includes all registered users as a struct
*/
list<Pointer<User> > DatabaseAccess::listUsers(){
   return _api.ref().getUsers();
}

/** removes all existing users from the gwen database */
void DatabaseAccess::deleteAllBanks(){

    list<Pointer<Bank> > banks = _api.ref().banks();
    list<Pointer<Bank> > :: iterator bit;

    for (bit = banks.begin(); bit != banks.end(); bit++)
       _api.ref().removeBank((*bit));   
}


/** Method to change used hbci version
*   checks if an overhanded hbci version is supported
*   by the bank and returns false if not
*/
bool DatabaseAccess::changeHbciVersion(int version){
  
    if (version != 201 && version != 210 && version != 220) 
         return false;
    else {
        list<Pointer<Bank> > banks = _api.ref().banks();
        Pointer <Bank> bank = *(banks.begin());

        list <int> versions = bank.ref().hbciVersions();
        list <int>::iterator it;

         if (versions.empty()) {
                     bank.ref().setHbciVersion(version);
                     return true;
         }

         for (it = versions.begin(); it != versions.end(); it++){
           if ((*it) == version) {
              bank.ref().setHbciVersion(version);
              return true;
           }
         } //end for

     } // end else       
   return false;
 }


 /** lists all transactions of an account which were made during
 *  a specific time periode */
 list<Pointer <Transaction> >  DatabaseAccess::listTransactions(int country,
                                                                const string &bankCode,
                                                                const string &accountId,
                                                                const string &accountSubId,
                                                                const Date &fromDate,
                                                                const Date &toDate){
  list<Pointer<Account> > accs;
  list<Pointer<Account> >::iterator ait;
  Pointer<Bank> bank;
  Pointer<Transaction> transaction;
  list<Pointer<Transaction > > result;
  Error err;
  GWEN_DB_NODE *xas;
  //bool first = true;

  accs =_api.ref().getAccounts(country, bankCode, accountId, accountSubId);

  if (accs.empty()) {
      cerr << "\nNo accounts found\n";
      return result;
  }
      
  for (ait=accs.begin(); ait!=accs.end(); ait++) {

     bank = _api.ref().findBank((*ait).ref().country(),
                                (*ait).ref().serverBankCode());
     if (!bank.isValid()){ 
        cerr << "\nInvalid Bank\n";
      return result;
     }
     xas=  (*ait).ref().storage().ref().getTransactions(MONEYPENNY_DEFAULT_COUNTRY,
                                                        "",
                                                        "",
                                                        fromDate,
                                                        toDate,
                                                        HOBIT_DUMP_FLAGS_DEFAULT);
     // if we got transactions
     if (xas) {
       GWEN_DB_NODE *n;
       n=GWEN_DB_GetFirstGroup(xas);
       if (n) {
                 cerr << "\n Transactions  found\n";
                 while(n) {
                   // get a single transaction
                    if (strcasecmp(GWEN_DB_GroupName(n), "transaction")==0) {
                      DBG_NOTICE(0, "Adding transaction "); 
                      transaction  = new Transaction(n); 
                      result.push_back(transaction);  
                    } // if transaction
                   n= GWEN_DB_GetNextGroup(n);
                 } // while n
             } // if (n)
       else return list<Pointer<Transaction > >();
       GWEN_DB_Group_free(xas);
     } // if transactions
 }  // end for
  return result;
}



/** lists all transactions of an account which were made during
 *  a specific time periode */
 list <Pointer <DatedTransfer> >  DatabaseAccess::listDatedTransfers(int country,
                                                                     const string &bankCode,
                                                                     const string &accountId,
                                                                     const string &accountSubId){

  list<Pointer<Account> > accs;
  list<Pointer<Account> >::iterator ait;
  Pointer<Bank> bank;
  Pointer<DatedTransfer> dt;
  list<Pointer<DatedTransfer > > result;
  Error err;
  GWEN_DB_NODE *xas;

   accs =_api.ref().getAccounts(country, bankCode, accountId, accountSubId);

    if (accs.empty()) {
     return result;
     cerr << "\nNo accounts found\n";
      }

  for (ait=accs.begin(); ait!=accs.end(); ait++) {
    xas=  (*ait).ref().getDatedTransfers();
    // if we got transactions
    if (xas) {
       GWEN_DB_NODE *n;
       n = GWEN_DB_GetFirstGroup(xas);
       if (n) {
          cerr << "\nFound dated transfers\n";
          while(n) {
             // get a single transaction
              if (strcasecmp(GWEN_DB_GroupName(n), "datedTransfer")==0) {
                 DBG_NOTICE(0, "Adding dated transaction ");
                 cerr << "Adding dated transaction ";
                 dt = new DatedTransfer(n);
                 result.push_back(dt);
              } // if transaction
             n=GWEN_DB_GetNextGroup(n);                                
          } // while n
       } // if (n)
       else cerr << "\nNo dated transfers found\n";
                     
      GWEN_DB_Group_free(n);
    } // if xas

 }  // end for

  return result;
}


/** lists all transactions of an account which were made during
 *  a specific time periode */
 list <Pointer <StandingOrder> >  DatabaseAccess::listStandingOrders(int country,
                                                                     const string &bankCode,
                                                                     const string &accountId,
                                                                     const string &accountSubId){

  list<Pointer<Account> > accs;
  list<Pointer<Account> >::iterator ait;
  Pointer<Bank> bank;
  Pointer<StandingOrder> so;
  list<Pointer<StandingOrder > > result;
  Error err;
  GWEN_DB_NODE *xas;

  accs =_api.ref().getAccounts(country, bankCode, accountId, accountSubId);

  if (accs.empty()) {
   return result;
   cerr << "No accounts found\n";
  }

  for (ait=accs.begin(); ait!=accs.end(); ait++) {
    xas=  (*ait).ref().getStandingOrders();
      // if we got transactions
      if (xas) {
        cerr << "\nFound standing orders\n";
        GWEN_DB_NODE *n;
        n=GWEN_DB_GetFirstGroup(xas);
        if (n) {
                     while(n) {
                    // get a single transaction
                     if (strcasecmp(GWEN_DB_GroupName(n), "standingOrder")==0) {
                     DBG_NOTICE(0, "Adding standing order ");
                     cerr << "Adding standing order ";
                     so = new StandingOrder(n);
                     result.push_back(so);
  
                     } // if transaction
                    n=GWEN_DB_GetNextGroup(n);
                } // while n
              } // if (n)
        else  cerr << "\nNo standing orders found\n";                       
        GWEN_DB_Group_free(n);
     } // if xas
 }  // end for
 return result;
}



/** This method generates the customers ini-letter */
string DatabaseAccess::generateIniLetter(int country, const string& bankcode, const string& customerId){

  Error err;
  list<Pointer<Customer> > custs;
  Pointer<Customer> cust;
  Pointer<Medium> medium;
  Pointer<MediumRDHBase> mediumrdh;
  Pointer<User> user;
  Pointer<Bank> bank;
  string modulus;
  string exponent;
  string myHash;
  int keyversion;
  int keynumber;
  QString result;
  bool isuser;
  string userid;
  string tmp;
  QString sysid = "MoneyPenny " + QString(VERSION);
  QDateTime date = QDateTime(QDateTime::currentDateTime());
  QString dateFormat = tr("yyyy/ddd/MMMM/d", "Date format for printer output");
  QString timeFormat = tr("hh:mm:ss ap", "Time format for printer output");
 
  custs=_api.ref().getCustomers(country, bankcode, customerId);
  if (custs.empty()) {
    fprintf(stderr,
            "%s",
            tr("Error: No matching customers found.\n").latin1());
    return "Error: No matching customers found";
  }
  if (custs.size()>1) {
    fprintf(stderr,
            "%s",
            tr("Error: Ambiguous customer specification.\n").latin1());
    return "Error: Ambiguous customer specification. ";
  }
  cust=custs.front();
  user=cust.ref().user();
  medium=user.ref().medium();
  bank=user.ref().bank();
  if (medium.ref().securityMode()!=HBCI_SECURITY_RDH) {
    fprintf(stderr,
            "%s",
            tr("Error: Not in RDH mode.\n").latin1());
    return "Error: Not in RDH mode.\n";
  }
  mediumrdh=medium.cast<MediumRDHBase>();

  // mount medium
  err=mediumrdh.ref().mountMedium();
  if (!err.isOk()) {
    fprintf(stderr,
            tr("Error mounting medium (%s)\n").latin1(),
            err.errorString().c_str());
    return "Error while trying to mount the security medium";
  }

  // select context
  err=mediumrdh.ref().selectContext(bank.ref().country(),
                                    bank.ref().bankCode(),
                                    user.ref().userId());
  if (!err.isOk()) {
    mediumrdh.ref().unmountMedium();
    fprintf(stderr,
            tr("Error selecting user context (%s)\n").latin1(),
            err.errorString().c_str());
      string error = "Error selecting user context";
      error += err.errorString().c_str();
      return error;
  }

  try {
    if (!mediumrdh.ref().userPubSignKey().isValid()) {
      fprintf(stderr,
              "%s",
              tr("Error: No sign key, please create keys.\n").latin1());
      return "Error: No sign key, please create keys.";
    }
    if (!mediumrdh.ref().userPubCryptKey().isValid()) {
      fprintf(stderr,
              "%s",
              tr("Error: No HBCI key, please create keys.\n").latin1());
      return "Error: No HBCI key, please create keys.";
    }

      // change the following lines to generate a bank ini letter:
      myHash=mediumrdh.ref().getUserIniLetterHash();
      exponent=mediumrdh.ref().getUserIniLetterExponent();
      modulus=mediumrdh.ref().getUserIniLetterModulus();
      keynumber=mediumrdh.ref().getUserKeyNumber();
      keyversion=mediumrdh.ref().getUserKeyVersion();
      isuser=true;
      userid=user.ref().userId();
      // end of change block


     // now we have all we need, unmount medium
    medium.ref().unmountMedium();

   result=  tr("INI-Letter generated by ") + sysid + "\n\n";
   result+= tr("Username             : __________________________\n");
   // xgettext:no-c-format
   result+= tr("Print Date           : ") + date.toString(dateFormat) + "\n";
   result+= tr("Time                 : ") + date.toString(timeFormat) + "\n";
   result+= tr("User                 : ")+ userid.c_str() + "\n";
   result+= tr("Key number           : ")+HBCI::String::num2string(keynumber).c_str() + "\n";
   result+= tr("Key version          : ")+HBCI::String::num2string(keyversion).c_str() + "\n";
   result+= tr("Customer system      : ")+sysid+"\n";
   result+="\n";

    result+=tr("My public key for electronic signature\n");
    result+=tr("  Exponent\n");

        if (exponent.length()<192)
            throw HBCI::Error("makeTxtLetter()",
                          ERROR_LEVEL_NORMAL,
                          0,
                          ERROR_ADVISE_DONTKNOW,
                          "exponent too short");
        result+= HoBIT_String::dumpHexString(exponent).c_str();

        result+="\n";
        result+=tr("  Modulus\n");
        if (exponent.length()<192)
            throw HBCI::Error("makeTxtLetter()",
                          ERROR_LEVEL_NORMAL,
                          0,
                          ERROR_ADVISE_DONTKNOW,
                          "exponent too short");
        result+= HoBIT_String::dumpHexString(modulus).c_str();

        result+="\n";
        result+=tr("  Hash value for my public key:\n");
        result+= HoBIT_String::dumpHexString(myHash).c_str();

        result+="\n";
        result+=tr("I confirm that my computer uses the above key for my electronic \nsignature.\n\n");
        result+=tr("I furthermore confirm that my computer received a bank public key \nwith the following hash value:\n");
        string bankHash = getBankFingerPrint();
        result+= QString(bankHash.c_str());
        result+="\n\n";

        result+=tr("_________________________  ____________________________\n"
                   "Location, date                            Signature\n");

 }
    catch (Error xerr) {
    DBG_ERROR(0, "Error: %s", xerr.errorString().c_str());
    string error = "Error: ";
    error += xerr.errorString().c_str();
    return error;
  }

  return result.latin1();
}

/** Returns the date of the last stored transaction.
  * This is for example needed for updating
  * transactions list, but can also be used for gui stuff
  */
Date DatabaseAccess::getLastDateOfTransactions(){

    Error err;
    list <Pointer<Account> > accs;
    list <Pointer<Account> >::iterator accit;

     int lastyear = 1980, lastmonth = 1, lastday = 1;
     accs =_api.ref().getAccounts();

    for (accit = accs.begin(); accit != accs.end(); accit++){
      int day, month, year;
       err=(*accit).ref().storage().ref().getLastDay(year, month, day);

       if (err.isOk())
            if (year >= lastyear)
              if (month >= lastmonth)
                if (day > lastday){
                        lastday = day;
                        lastmonth = month;
                        lastyear = year;
                }
                    
    }   // end for

    return Date(lastday, lastmonth, lastyear);
}


/**
* Method that reads out the ZKA -banklist fints_institute.csv and returns 
* a list of HoBIT_Bank-Objects
*
*/
list<HoBIT_Bank> DatabaseAccess::readBankFile(const string &filename) 
{
   // First create list
    HoBIT_Bank * bank;
    list<HoBIT_Bank> banks;
    QFile file(filename.c_str());
    QFileInfo fileInfo(file);
    qDebug("Reading bankfile: %s", filename.c_str());
    if ( file.open( IO_ReadOnly ) ) {
        QTextStream stream(&file);
        int nLine = 1;
        QString line = stream.readLine(); // read headerLine
        nLine ++;
        if(fileInfo.extension() =="txt"){
          stream.setEncoding(QTextStream::Latin1);
          if(line.find("10000000") == 0)
          while ( !stream.atEnd() ) {           
              line = stream.readLine(); // line of text excluding '\n'
              bank = new HoBIT_Bank();
              if(bank->setFromTxtLine(line))
                banks.push_back(*bank);
              else delete bank; //invalid line
              nLine++; 
          }
          else cerr << "File format error in bank data file: " <<filename;
        }
        else if(fileInfo.extension() =="csv"){
          stream.setEncoding(QTextStream::Latin1);
          if(line.find("Nr.;") == 0)
          while ( !stream.atEnd() ) {           
              line = stream.readLine(); // line of text excluding '\n'
              bank = new HoBIT_Bank();
              if(bank->setFromCsvLine(line))
                banks.push_back(*bank);
              else delete bank; //invalid line
              nLine++; 
          }
          else cerr << "File format error in bank data file: " <<filename;
        }
        file.close();
    }//end if fileOpen
    else cerr << "Bank data file not found: " << filename << endl;
   return banks;
}


/** returns the hash fingerprint of the institute after resceiving the bank servers public keys */
string DatabaseAccess::getBankFingerPrint(){

Error err;
  list<Pointer<Customer> > custs;
  Pointer<Customer> cust;
  Pointer<Medium> medium;
  Pointer<MediumRDHBase> mediumrdh;
  Pointer<User> user;
  Pointer<Bank> bank;
  string modulus;
  string exponent;
  string hash;
  int keyversion;
  int keynumber;
  string result;
  bool isuser;
  string userid;
  string tmp;
  bool usecrypt;

  custs=_api.ref().getCustomers();
      
  if (custs.empty()) {
    fprintf(stderr,
            "%s",
            tr("Error: No matching customers found.\n").latin1());
    return "Error: No matching customers found";
  }
  
  if (custs.size()>1) {
    fprintf(stderr,
            "%s",
            tr("Error: Ambiguous customer specification.\n").latin1());
    return "Error: Ambiguous customer specification. ";
  }
  
  cust=custs.front();
  user=cust.ref().user();
  medium=user.ref().medium();
  bank=user.ref().bank();
  if (medium.ref().securityMode()!=HBCI_SECURITY_RDH) {
    fprintf(stderr,
            "%s",
            tr("Error: Not in RDH mode.\n").latin1());
    return "Error: Not in RDH mode.\n";
  }
  
  mediumrdh=medium.cast<MediumRDHBase>();

  // mount medium
  err=mediumrdh.ref().mountMedium();
  if (!err.isOk()) {
    fprintf(stderr,
            tr("Error mounting medium (%s)\n").latin1(),
            err.errorString().c_str());
    return "Error while trying to mount the security medium";
  }

  // select context
  err=mediumrdh.ref().selectContext(bank.ref().country(),
                                    bank.ref().bankCode(),
                                    user.ref().userId());
  if (!err.isOk()) {
    mediumrdh.ref().unmountMedium();
    fprintf(stderr,
            tr("Error selecting user context (%s)\n").latin1(),
            err.errorString().c_str());
      string error = "Error selecting user context";
      error += err.errorString().c_str();
      return error;
  }

  try {
    if (!mediumrdh.ref().userPubSignKey().isValid()) {
      fprintf(stderr,
              "%s",
              tr("Error: No sign key, please create keys.\n").latin1());
      return "Error: No sign key, please create keys.";
    }
    if (!mediumrdh.ref().userPubCryptKey().isValid()) {
      fprintf(stderr,
              "%s",
              tr("Error: No crypt key, please create keys.\n").latin1());
      return "Error: No crypt key, please create keys.";
    }

      // change the following lines to generate a users ini letter:
      usecrypt=!mediumrdh.ref().hasInstSignKey();
      hash=mediumrdh.ref().getInstIniLetterHash(usecrypt);
      exponent=mediumrdh.ref().getInstIniLetterExponent(usecrypt);
      modulus=mediumrdh.ref().getInstIniLetterModulus(usecrypt);
      keynumber=-1;
      keyversion=-1;
      isuser=false;
      userid=bank.ref().bankCode();
      // end of change block

      mediumrdh=medium.cast<MediumRDHBase>();

      result=HoBIT_String::dumpHexString(hash);    
  }

      catch (Error xerr) {
    DBG_ERROR(0, "Error: %s", xerr.errorString().c_str());
    string error = "Error: ";
    error += xerr.errorString().c_str();
    return error;
  }

  return result;
} 



/** returns the hash fingerprint of the own public key 
*/
string DatabaseAccess::getMyOwnFingerPrint(){

Error err;
  list<Pointer<Customer> > custs;
  Pointer<Customer> cust;
  Pointer<Medium> medium;
  Pointer<MediumRDHBase> mediumrdh;
  Pointer<User> user;
  Pointer<Bank> bank;
  string modulus;
  string exponent;
  string hash;
  int keyversion;
  int keynumber;
  string result;
  bool isuser;
  string userid;
  string tmp;

  custs=_api.ref().getCustomers();

  if (custs.empty()) {
    fprintf(stderr,
            "%s",
            tr("Error: No matching customers found.\n").latin1());
    return "Error: No matching customers found";
  }

  if (custs.size()>1) {
    fprintf(stderr,
            "%s",
            tr("Error: Ambiguous customer specification.\n").latin1());
    return "Error: Ambiguous customer specification. ";
  }

  cust=custs.front();
  user=cust.ref().user();
  medium=user.ref().medium();
  bank=user.ref().bank();
  if (medium.ref().securityMode()!=HBCI_SECURITY_RDH) {
    fprintf(stderr,
            "%s",
            tr("Error: Not in RDH mode.\n").latin1());
    return "Error: Not in RDH mode.\n";
  }

  mediumrdh=medium.cast<MediumRDHBase>();

  // mount medium
  err=mediumrdh.ref().mountMedium();
  if (!err.isOk()) {
    fprintf(stderr,
            tr("Error mounting medium (%s)\n").latin1(),
            err.errorString().c_str());
    return "Error while trying to mount the security medium";
  }

  // select context
  err=mediumrdh.ref().selectContext(bank.ref().country(),
                                    bank.ref().bankCode(),
                                    user.ref().userId());
  if (!err.isOk()) {
    mediumrdh.ref().unmountMedium();
    fprintf(stderr,
            tr("Error selecting user context (%s)\n").latin1(),
            err.errorString().c_str());
      string error = "Error selecting user context";
      error += err.errorString().c_str();
      return error;
  }

  try {
    if (!mediumrdh.ref().userPubSignKey().isValid()) {
      fprintf(stderr,
              "%s",
              tr("Error: No sign key, please create keys.\n").latin1());
      return "Error: No sign key, please create keys.";
    }
    if (!mediumrdh.ref().userPubCryptKey().isValid()) {
      fprintf(stderr,
              "%s",
              tr("Error: No crypt key, please create keys.\n").latin1());
      return "Error: No crypt key, please create keys.";
    }

      // change the following lines to generate a bank hash:
      hash=mediumrdh.ref().getUserIniLetterHash();
      exponent=mediumrdh.ref().getUserIniLetterExponent();
      modulus=mediumrdh.ref().getUserIniLetterModulus();
      keynumber=mediumrdh.ref().getUserKeyNumber();
      keyversion=mediumrdh.ref().getUserKeyVersion();
      isuser=true;
      userid=user.ref().userId();
      // end of change block

      mediumrdh=medium.cast<MediumRDHBase>();

      result=HoBIT_String::dumpHexString(hash);
  }

      catch (Error xerr) {
    DBG_ERROR(0, "Error: %s", xerr.errorString().c_str());
    string error = "Error: ";
    error += xerr.errorString().c_str();
    return error;
  }

  return result;
} 



 /** changes the account description */
void DatabaseAccess::changeAccountName(const string& accid, const string& name){
    list<Pointer<Account> > accs = _api.ref().getAccounts(0, "*", accid, "*");
    list<Pointer<Account> >:: iterator accit;
       
    for (accit = accs.begin(); accit != accs.end(); accit++)
      (*accit).ref().setAccountName(name);    
}



/** creates a new account with the passed values
  * this procedure will for example be executed,  
  * the retrieving of the account list from the 
  * bank server fails 
  *
  * @param name: Name of the account owner
  * @param accountId: the account number
  * @param accountSubId: Hbci allows subids for accounts 
  * 			 this parameter is normaly not needed
  * @param accountName:  you can give your account an description
  *
  */		
HobitError DatabaseAccess::addAccount(const string& name, const string &accountId, const string &accountSubId, const string& accountName){

  // variables declaration
  Error err;
  list<Pointer<Customer> > custs;
  Pointer<Customer> cust;
  Pointer<Bank> bank;
  Pointer<Account> account;
  string bankCode;


  custs=_api.ref().getCustomers();
  
  if (custs.empty()) {
    fprintf(stderr,
            "%s",
            "No matching customers found.\n");
    return HobitErrorInvalidArgs;
  }

  if (custs.size()>1) {
    fprintf(stderr,
	    "%s",
	    "Ambiguous customer specification.\n");
    return HobitErrorInvalidArgs;
  }

  try {
    cust=custs.front();

    bank=cust.ref().user().ref().bank();
    
    
    account=
        _api.ref().accountFactory(cust.ref().custId(),
                         	  name,
                            	  accountId,
                             	  accountSubId,
                            	  bank.ref().country(),
                            	  bank.ref().bankCode(),
                            	  bank.ref().bankCode(),
                            	  accountName);
				  
    err=_api.ref().addAccount(account);
    
    if (!err.isOk()) {
      DBG_ERROR(0, "Error: %s", err.errorString().c_str());
      return HobitErrorExecution;
    }

    fprintf(stdout,
	    tr("Added account %d/%s/%s\n"),
	    bank.ref().country(),
	    bankCode.c_str(),
	    accountId.c_str());

    
  } // try
  
  catch (Error xerr) {
    DBG_ERROR(0, "Error: %s", xerr.errorString().c_str());
    return HobitErrorExecution;
  }

  return HobitErrorSuccess;


}

  /** writes all templates to file */
  Error DatabaseAccess::saveTransferTemplates(const string& templateFile){
    	Error err;
  	GWEN_DB_NODE *db;
  	GWEN_DB_NODE *n;
  	list<Pointer<HoBIT_TransferTemplate> >::iterator tit;
	
	db=GWEN_DB_Group_new("DatabaseAccess::saveTransferTemplates");
	cerr << "Now Saving Templates";
	
	try {
    		DBG_DEBUG(0, "Saving accounts");
    		for (tit=_templates.begin(); tit!=_templates.end(); tit++) {
      
      		n=GWEN_DB_GetGroup(db,
                	           GWEN_DB_FLAGS_DEFAULT |
                        	   GWEN_PATH_FLAGS_CREATE_GROUP,
                            	  "transfer_templates/template");
      		assert(n);
      		(*tit).ref().toConfig(n);
    		} // for
    		DBG_DEBUG(0, "Saving transfer templates done");	
	} // try	
	
	catch (Error xerr) {
   	GWEN_DB_Group_free(db);
	cerr << "Saving Templates failed";
    	return Error("DatabaseAccess::saveTransferTemplates", xerr);
	}
	
	
	 DBG_DEBUG(0, "Saving Templates file \"%s\"",
            	   templateFile.c_str());
         
	 if (GWEN_DB_WriteFile(db, templateFile.c_str(),
         	               GWEN_DB_FLAGS_DEFAULT)) {
    	GWEN_DB_Group_free(db);
    	return Error("DatabaseAccess::saveTrasferTemplates",
        	     ERROR_LEVEL_NORMAL,
                     HBCI_ERROR_CODE_UNKNOWN,
                     ERROR_ADVISE_DONTKNOW,
                    "Error saving Templates config file",
                     templateFile);
  }

  GWEN_DB_Group_free(db);
  DBG_DEBUG(0, "Saving templates done.");
  return Error();
	
  }
	
	
  
  /** reads stored templates from file */
  Error DatabaseAccess::readTransferTemplates(const string& templateFile){
  
  Error err;
  GWEN_DB_NODE *db;
  GWEN_DB_NODE *n;
  GWEN_DB_NODE *gr;

  DBG_INFO(0, "Reading templates");
  
  
  if (access(templateFile.c_str(), F_OK)) {
    DBG_WARN(0,
             "Template file \"%s\" does not exist, will create it later",
              templateFile.c_str());
  }
  else {
    DBG_INFO(0, "Loading Templates file \"%s\"",
             templateFile.c_str());
    
    db=GWEN_DB_Group_new("DatabaseAccess::readTransferTemplates");
    if (GWEN_DB_ReadFile(db, templateFile.c_str(),
			 GWEN_DB_FLAGS_DEFAULT|
			 GWEN_PATH_FLAGS_CREATE_GROUP)) {
      GWEN_DB_Group_free(db);
      return Error("DatabaseAccess::readTransferTemplates",
		   ERROR_LEVEL_NORMAL,
		   HBCI_ERROR_CODE_UNKNOWN,
		   ERROR_ADVISE_DONTKNOW,
		   "Error loading Template config file",
		   templateFile);
    }


  try {
  
  gr=GWEN_DB_GetGroup(db, GWEN_PATH_FLAGS_NAMEMUSTEXIST, "transfer_templates");
      if (gr) {
        n=GWEN_DB_GetFirstGroup(gr);
        while (n) {
          if (strcasecmp(GWEN_DB_GroupName(n), "template")==0) {
            // load templates
            Pointer<HoBIT_TransferTemplate> temp;

            temp=new HoBIT_TransferTemplate(n);
            addTransferTemplate(temp);
	    
          } // if account
          else {
            DBG_INFO(0, "Unkown section \"%s\", ignoring",
                     GWEN_DB_GroupName(n));
          }
          n=GWEN_DB_GetNextGroup(n);
        } // while n
      } // if "templates"
  
  
      }	catch (Error xerr) {
        	GWEN_DB_Group_free(db);
      		return Error("DatabaseAccess::readTransferTemplates", xerr);
    }
    GWEN_DB_Group_free(db);
  }
  DBG_DEBUG(0, "Reading templates done.");
  return Error();

}



  /** adding transfer template */
  void DatabaseAccess::addTransferTemplate(Pointer<HoBIT_TransferTemplate> temp){
  _templates.push_back(temp);
  }

  
  
  
/** deletes an transfer template */  
void DatabaseAccess::deleteTransferTemplate(   	const string& ownAccountId,
  			      			const string& otherName, 
			      			const string& otherAccountId,
			      			const string& otherBank,
			      			const string& value,
			      			const string& purpose1,
			      			const string& purpose2) {
  Error err;
  list<Pointer<HoBIT_TransferTemplate> >::iterator tit;
  
  for (tit = _templates.begin(); tit != _templates.end(); tit++){
  
  if (	ownAccountId == (*tit).ref().ownAccountId() &&
  	otherName == (*tit).ref().otherName() &&
  	otherAccountId == (*tit).ref().otherAccountId() &&
	value == (*tit).ref().value().toReadableString() &&
	otherBank == (*tit).ref().otherBank() &&
	purpose1 == (*tit).ref().purpose1() &&
	purpose2 == (*tit).ref().purpose2()){
   	_templates.remove(*tit);
	break;
   }
  }
}
